#ifndef TrackChecker_h
#define TrackChecker_h 1

#include "marlin/Processor.h"

#include "lcio.h"

#include <string>
#include <vector>
#include <map>

#include <EVENT/LCCollection.h>
#include <EVENT/Track.h>

#include <AIDA/AIDA.h>
#include "TH1F.h"
#include "TTree.h"
#include "TFile.h"

class TTree;
class TCanvas;

using namespace lcio ;
using namespace marlin ;
using namespace AIDA ;

class Global_residuals : public Processor {
		
 public:
	
  virtual Processor*  newProcessor() { return new Global_residuals ; }
	
  Global_residuals() ;
	
  // Initialisation - run at the beginning to start histograms, etc.
  virtual void init() ;
	
  // Called at the beginning of every run
  virtual void processRunHeader( LCRunHeader* run ) ;
	
  // Run over each event - the main algorithm
  virtual void processEvent( LCEvent * evt ) ;
	
  // Run at the end of each event
  virtual void check( LCEvent * evt ) ;
	
  // Called at the very end for cleanup, histogram saving, etc.
  virtual void end() ;
	
  // Call to get collections
  void getCollection(LCCollection*&, std::string, LCEvent*);
	
  void angleInFixedRange(double& angle);

  void clearEventVariables();

  double getDistanceFromClosestHit(Track*& trackA, Track*& trackB);

  double getDist(TrackerHit*& hitA, TrackerHit*& hitB);



 protected:
	
  // Collection names for (in/out)put
  std::string m_inputTrackCollection ;
  std::string m_inputTrackRelationCollection ;
  std::string m_inputParticleCollection ;
 	
  
  std::string _outFileNameG;
  std::string _outFileTypeG;
        std::string _treeName;
  // Run and event counters
  int m_eventNumber ;
  int m_runNumber ;
  
  // Magnetic field
  float m_magneticField;
	
  //Flags
  bool m_useOnlyTree;

  
 TFile* _out;
 TTree* perftree ;
  
  // Output histograms



  TCanvas* pulls;
  TCanvas* res;

  TH1F* omegaPull;
  TH1F* phiPull;
  TH1F* tanLambdaPull;
  TH1F* d0Pull;
  TH1F* z0Pull ;

  TH1F*  OmegaResidual;
  TH1F* PhiResidual;
  TH1F* TanLambdaResidual;
  TH1F* d0Residual;
  TH1F* z0Residual;
  
  TH1F* m_omegaMCParticle;
  TH1F* m_phiMCParticle;
  TH1F* m_tanLambdaMCParticle;
  TH1F* m_d0MCParticle;
  TH1F* m_z0MCParticle;

  TH1F* m_omegaTrack;
  TH1F* m_phiTrack;
  TH1F* m_tanLambdaTrack;
  TH1F* m_d0Track;
  TH1F* m_z0Track;
  
  TH1F* m_trackChi2;
  

  std::string m_treeName;

  

  std::vector<double > truePt;
  std::vector<double > trueTheta;
  std::vector<double > trueEta;
  std::vector<double > truePhi;
  std::vector<double > trueD0;
  std::vector<double > trueZ0;
  std::vector<double > trueP;

  std::vector<double > recoPt;
  std::vector<double > recoTheta;
  std::vector<double > recoEta;
  std::vector<double > recoPhi;
  std::vector<double > recoD0;
  std::vector<double > recoZ0;
  std::vector<double > recoP;

  std::vector<int > recoNhits;
  std::vector<double > recoChi2OverNDF;
  std::vector<double > recoMinDist;

  std::vector<double > pullOm;
  std::vector<double > pullPh;
  std::vector<double > pullTL;
  std::vector<double > pullD0;
  std::vector<double > pullZ0;
  std::vector<double > pullEta;
  std::vector<double > pullPt;

  std::vector<double > omegaResidual;
  std::vector<double > phiResidual;
  std::vector<double > resTanLambda;
  std::vector<double > resD0;
  std::vector<double > resZ0;
  std::vector<double > resEta;
  std::vector<double > resPt;
 // std::vector<double > resPt_err;
  
  //added  uncertainties
  std::vector<double > _d0trackError;
  std::vector<double > _phtrackError;
  std::vector<double > _omtrackError;
  std::vector<double > _z0trackError;
  std::vector<double > _tLtrackError;
  std::vector<double > _pttrackError;
} ;

#endif



