/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
#include "Global_residuals.h"
#include "LinkDef.h"

#include "DDRec/API/IDDecoder.h"

#include <marlinutil/HelixClass.h>
#include "MarlinTrk/MarlinTrkUtils.h"
#include "MarlinTrk/HelixTrack.h"
#include "MarlinTrk/HelixFit.h"
#include "MarlinTrk/HelixFit.h"
#include "MarlinTrk/IMarlinTrack.h"
#include "MarlinTrk/Factory.h"
#include "MarlinTrk/MarlinTrkDiagnostics.h"
#include "MarlinTrk/IMarlinTrkSystem.h"


#include <EVENT/LCCollection.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/LCRelationImpl.h>
#include <EVENT/SimTrackerHit.h>
#include <IMPL/TrackerHitPlaneImpl.h>
#include <EVENT/MCParticle.h>
#include <IMPL/TrackImpl.h>

#include <UTIL/CellIDEncoder.h>
#include <UTIL/ILDConf.h>
#include <UTIL/BitSet32.h>
#include <UTIL/LCRelationNavigator.h>

#include "DD4hep/LCDD.h"
#include "DD4hep/DD4hepUnits.h"
#include "DDRec/SurfaceManager.h"

#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>

#include "marlin/ProcessorEventSeeder.h"
#include "marlin/Global.h"

#include "CLHEP/Vector/TwoVector.h"

#include <AIDA/IAnalysisFactory.h>
#include <AIDA/IHistogramFactory.h>

#include <cmath>
#include <math.h>
#include <algorithm>
#include <sstream>
#include <iostream>
#include <climits>
#include <cfloat>
#include "TH1F.h"
#include "TCanvas.h"
#include "TFile.h"
#include "TStyle.h"
#include "TTree.h"

using namespace lcio ;
using namespace marlin ;
using namespace std ;
using namespace DD4hep ;
using namespace AIDA ;
using namespace std ;

Global_residuals aGlobal_residuals;
/*

  Generic code set up to quickly add algorithms
 
*/
Global_residuals::Global_residuals() : Processor("Global_residuals"){
  
_description = "TrackChecker plots pull distributions, track resolution parameters and reconstruction efficiencies" ;
registerInputCollection( LCIO::TRACK,
                           "TrackCollectionName",
                           "Track collection name",
                           m_inputTrackCollection,

			std::string("SiTracks") );
			
			 registerInputCollection( LCIO::LCRELATION,
					"TrackRelationCollectionName",
                           "Track relation collection name",
                           m_inputTrackRelationCollection,	  
			 std::string("SiTrackRelations"));

			 
			   registerInputCollection( LCIO::MCPARTICLE,
                           "MCParticleCollectionName",
                           "Name of the MCParticle input collection",
                           m_inputParticleCollection,
                           std::string("MCParticle"));
			   
			     registerProcessorParameter("UseOnlyTree",
                             "Flag to fill only tree variables or also histogram version of pulls and residuals",
                             m_useOnlyTree,
                             bool(true));
			     
			     registerProcessorParameter( "outFileNameG",
                              "Name of the output root file",
                              _outFileNameG,
                              std::string()
                              );		
			  
			     registerProcessorParameter( "outFileTypeG",
                              "Name of the output root file",
                              _outFileTypeG,
                              std::string("root")
                              );	     
}

void Global_residuals::init(){
// Print the initial parameters
	printParameters() ;
	
	// Reset counters
	m_runNumber = 0 ;
	m_eventNumber = 0 ;
  
  // Get the magnetic field
  DD4hep::Geometry::LCDD& lcdd = DD4hep::Geometry::LCDD::getInstance();
  const double position[3]={0,0,0}; // position to calculate magnetic field at (the origin in this case)
  double magneticFieldVector[3]={0,0,0}; // initialise object to hold magnetic field
  lcdd.field().magneticField(position,magneticFieldVector); // get the magnetic field vector from DD4hep
  m_magneticField = magneticFieldVector[2]/dd4hep::tesla; // z component at (0,0,0)

	// Register this process
	Global::EVENTSEEDER->registerProcessor(this);
}

void Global_residuals::processRunHeader( LCRunHeader* run) {
	++m_runNumber ;
}

void Global_residuals::processEvent( LCEvent* evt ) {
 if( isFirstEvent() ) { 
   if(_outFileTypeG=="root")_outFileNameG+=".root";
     
    _out = new TFile(_outFileNameG.c_str(),"RECREATE");
	_treeName="t";
   
   perftree = new TTree(_treeName.c_str(),_treeName.c_str());//m_treeName m_treeName
    int bufsize = 32000; //default buffer size 32KB
          
    perftree->Branch("truePt","std::vector<double >",&truePt,bufsize,0) ;
    perftree->Branch("trueTheta","std::vector<double >",&trueTheta,bufsize,0) ;
    perftree->Branch("truePhi","std::vector<double >",&truePhi,bufsize,0) ;
    perftree->Branch("trueD0","std::vector<double >",&trueD0,bufsize,0) ;
    perftree->Branch("trueZ0","std::vector<double >",&trueZ0,bufsize,0) ;
    perftree->Branch("trueP","std::vector<double >",&trueP,bufsize,0) ;
    perftree->Branch("trueEta","std::vector<double >",&trueEta,bufsize,0) ;

    perftree->Branch("recoPt","std::vector<double >",&recoPt,bufsize,0) ;
    perftree->Branch("recoTheta","std::vector<double >",&recoTheta,bufsize,0) ;
    perftree->Branch("recoPhi","std::vector<double >",&recoPhi,bufsize,0) ;
    perftree->Branch("recoD0","std::vector<double >",&recoD0,bufsize,0) ;
    perftree->Branch("recoZ0","std::vector<double >",&recoZ0,bufsize,0) ;
    perftree->Branch("recoP","std::vector<double >",&recoP,bufsize,0) ;
    perftree->Branch("recoEta","std::vector<double >",&recoEta,bufsize,0) ;
 



      perftree->Branch("pullOm","std::vector<double >",&pullOm,bufsize,0) ;
      perftree->Branch("pullPh","std::vector<double >",&pullPh,bufsize,0) ;
      perftree->Branch("pullTL","std::vector<double >",&pullTL,bufsize,0) ;
      perftree->Branch("pullD0","std::vector<double >",&pullD0,bufsize,0) ;
      perftree->Branch("pullZ0","std::vector<double >",&pullZ0,bufsize,0) ;
      //perftree->Branch("pullPt","std::vector<double >",&pullPt,bufsize,0) ;
      perftree->Branch("pullPt","std::vector<double >",&pullPt,bufsize,0) ; //resPt_err
      
      
      perftree->Branch("_d0trackError","std::vector<double >",&_d0trackError,bufsize,0) ;
      perftree->Branch("_phtrackError","std::vector<double >",&_phtrackError,bufsize,0) ;
      perftree->Branch("_omtrackError ","std::vector<double >",&_omtrackError ,bufsize,0) ;
      perftree->Branch("_z0trackError","std::vector<double >",&_z0trackError,bufsize,0) ;
      perftree->Branch("_tLtrackError","std::vector<double >",&_tLtrackError,bufsize,0) ;
      perftree->Branch("_pttrackError","std::vector<double >",&_pttrackError,bufsize,0) ;
  


      
      
      perftree->Branch("resOmega","std::vector<double >",&omegaResidual,bufsize,0) ;
      perftree->Branch("resPhi","std::vector<double >",&phiResidual,bufsize,0) ;
      perftree->Branch("resTanLambda","std::vector<double >",&resTanLambda,bufsize,0) ;
      perftree->Branch("resD0","std::vector<double >",&resD0,bufsize,0) ;
      perftree->Branch("resZ0","std::vector<double >",&resZ0,bufsize,0) ;
      perftree->Branch("resPt","std::vector<double >",&resPt,bufsize,0) ;
   

      
 omegaPull = new TH1F("OmegaPull","Omega pull",100,-10,10);
    phiPull = new TH1F("PhiPull","Phi pull",100,-10,10);
    tanLambdaPull = new TH1F("TanLambdaPull","TanLambda",100,-10,10);
    d0Pull = new TH1F("d0pull","d0 pull",100,-10,10);
    z0Pull = new TH1F("z0pull","z0",100,-10,10); 
   
        OmegaResidual = new TH1F("OmegaResidual","Omega Residual",50,-0.001,0.001);
    PhiResidual = new TH1F("PhiResidual","Phi Residual",50,-0.005,0.005);
    TanLambdaResidual = new TH1F("TanLambdaResidual","TanLambda Residual",50,-0.005,0.005);
    d0Residual = new TH1F("d0Residual","d0 Residual",50,-0.1,0.1);
    z0Residual = new TH1F("z0Residual","z0 Residual",50,-0.1,0.1); 
    
     m_omegaMCParticle = new TH1F("OmegaMCParticle","OmegaMCParticle",200,-10,10);
      m_phiMCParticle = new TH1F("PhiMCParticle","PhiMCParticle",200,-10,10);
      m_tanLambdaMCParticle = new TH1F("TanLambdaMCParticle","TanLambdaMCParticle",200,-10,10);
      m_d0MCParticle = new TH1F("D0MCParticle","D0MCParticle",200,-10,10);
      m_z0MCParticle = new TH1F("Z0MCParticle","Z0MCParticle",200,-10,10);
  
      m_omegaTrack = new TH1F("OmegaTrack","OmegaTrack",200,-10,10);
      m_phiTrack = new TH1F("PhiTrack","PhiTrack",200,-10,10);
      m_tanLambdaTrack = new TH1F("TanLambdaTrack","TanLambdaTrack",200,-10,10);
      m_d0Track = new TH1F("D0Track","D0Track",200,-10,10);
      m_z0Track = new TH1F("Z0Track","Z0Track",200,-10,10);
  
      m_trackChi2 = new TH1F("TrackChi2","TrackChi2",500,0,10);
}
  clearEventVariables();
// Get the collection of tracks
  LCCollection* trackCollection = 0 ;
  getCollection(trackCollection, m_inputTrackCollection, evt); if(trackCollection == 0) return;

  // Get the collection of MC particles
  LCCollection* particleCollection = 0 ;
  getCollection(particleCollection, m_inputParticleCollection, evt); if(particleCollection == 0) return;

  // Get the collection of track relations to MC particles
  LCCollection* trackRelationCollection = 0 ;
  getCollection(trackRelationCollection, m_inputTrackRelationCollection, evt); if(trackRelationCollection == 0) return;
  
  // Create the relations navigator
  LCRelationNavigator* relation = new LCRelationNavigator( trackRelationCollection );
  
 /*
    Loop over all tracks, get the MC particle that produced it, and make
    a helical track fit to produce the "true" path of the particle. Then
    calculate the pull distributions and absolute error distributions.
  */
  
  // Loop over all tracks
  int nParticles= particleCollection->getNumberOfElements();
  
  for(int itParticle=0;itParticle<nParticles;itParticle++)
  {
       MCParticle* particle = dynamic_cast<MCParticle*>( particleCollection->getElementAt(itParticle) ) ;
   // get theta angle for plot: eff vs theta
   std::cout<<"These are particles PDG's  in events  "<<particle->getPDG()<<std::endl;
  

  
  
  
  int nTracks = trackCollection->getNumberOfElements();
for(int itTrack=0;itTrack<nTracks;itTrack++){

 // Get the track
    Track* track = dynamic_cast<Track*>( trackCollection->getElementAt(itTrack) ) ;
    
    // Get the related list of MC particles
    const LCObjectVec& mcparticleVector = relation->getRelatedToObjects( track );

    
    
    // Take the first MC particle (there should only be one)
  

    MCParticle* particle = dynamic_cast<MCParticle*>(mcparticleVector.at(0));
    
   
      std::cout<<"There is only one MCParticle  "<<std::endl;
    
    


    
     /*
      Now apply some cuts - only valid MC particles will be considered
    */
    
    //    if(particle->isDecayedInTracker()) continue; // Ignore tracks that decay inside the tracker volume
		
    
    // Construct the true helical trajectory
    HelixClass helix ;
    float pos[3]={particle->getVertex()[0],particle->getVertex()[1],particle->getVertex()[2]}; // Vertex position
    float mom[3]={particle->getMomentum()[0],particle->getMomentum()[1],particle->getMomentum()[2]}; // Particle momentum
    float charge = particle->getCharge() ; // Particle charge
    helix.Initialize_VP(pos,mom,charge,m_magneticField) ; // Initialised helix track

    // Now get the pull distributions
    double d0mcp = helix.getD0() ;
    double phmcp = helix.getPhi0() ;
    double ommcp = helix.getOmega() ;
    double z0mcp = helix.getZ0() ;
    double tLmcp = helix.getTanLambda() ;

    double d0track = track->getD0() ;
    double phtrack = track->getPhi() ;
    double omtrack = track->getOmega() ;
    double z0track = track->getZ0() ;
    double tLtrack = track->getTanLambda() ;

    double d0trackError = track->getCovMatrix()[0] ;
    double phtrackError = track->getCovMatrix()[2] ;
    double omtrackError = track->getCovMatrix()[5] ;
    double z0trackError = track->getCovMatrix()[9] ;
    double tLtrackError = track->getCovMatrix()[14] ;
  
   
    
   


    //angle conversions
    angleInFixedRange(phmcp);
	  angleInFixedRange(phtrack);    
    if (m_useOnlyTree) {
      
      
    omegaResidual.push_back(omtrack-ommcp);
      phiResidual.push_back(phtrack-phmcp);
      resTanLambda.push_back(tLtrack-tLmcp);
      resD0.push_back(d0track-d0mcp);
      resZ0.push_back(z0track-z0mcp);
     //resPt.push(truePt.back()-recoPt.back());
      
      
      
      pullOm.push_back(omegaResidual.back()/sqrt(omtrackError)); 
      pullPh.push_back(phiResidual.back()/sqrt(phtrackError)); 
      pullTL.push_back(resTanLambda.back()/sqrt(tLtrackError)); 
      pullD0.push_back(resD0.back()/sqrt(d0trackError)); 
      pullZ0.push_back(resZ0.back()/sqrt(z0trackError)); 
     
      
      _d0trackError.push_back(d0trackError );
      _phtrackError.push_back(phtrackError);
      _omtrackError.push_back(omtrackError);
      _z0trackError.push_back(z0trackError );
      _tLtrackError.push_back(tLtrackError);
      
      
      
      // pullPt.push_back();
      
      
    }
    
    else {
      // Fill the output histograms
     /* d0Pull->Fill( (d0track-d0mcp)/sqrt(d0trackError) );
      phiPull->Fill( (phtrack-phmcp)/sqrt(phtrackError) );
      omegaPull->Fill( (omtrack-ommcp)/sqrt(omtrackError) );
      z0Pull->Fill( (z0track-z0mcp)/sqrt(z0trackError) );
      tanLambdaPull->Fill( (tLtrack-tLmcp)/sqrt(tLtrackError) );
 */
  /*    resD0->Fill(d0track-d0mcp);
      phiResidual->Fill(phtrack-phmcp);
      omegaResidual->Fill(omtrack-ommcp);
      resZ0->Fill(z0track-z0mcp);
      resTanLambda->Fill(tLtrack-tLmcp);
*/
      m_omegaMCParticle->Fill(ommcp);
      m_phiMCParticle->Fill(phmcp);
      m_tanLambdaMCParticle->Fill(tLmcp);
      m_d0MCParticle->Fill(d0mcp);
      m_z0MCParticle->Fill(z0mcp);
    
      m_omegaTrack->Fill(omtrack);
      m_phiTrack->Fill(phtrack);
      m_tanLambdaTrack->Fill(tLtrack);
      m_d0Track->Fill(d0track);
      m_z0Track->Fill(z0track);
    
      m_trackChi2->Fill(track->getChi2());
      //    if(track->getChi2() < 0.05) std::cout<<"EVENT NUMBER "<<m_eventNumber<<std::endl;
    }
    
    
	  truePt.push_back(0.3*m_magneticField/fabs(ommcp*1000.)); //omega in 1/mm -> transformed in 1/m
    double trueThetaRad = M_PI/2 - atan(tLmcp); 
	  trueTheta.push_back(trueThetaRad*180./M_PI);
	  truePhi.push_back(phmcp*180./M_PI);
	  trueD0.push_back(d0mcp);
	  trueZ0.push_back(z0mcp);
	  trueP.push_back(truePt.back()/sin(trueThetaRad)); 
	  double tantrue = (double)tan((trueThetaRad*180./M_PI)/2);
	   double _trueEta = std::log(tantrue);
          trueEta.push_back(_trueEta);
	  
	  
	  recoPt.push_back(0.3*m_magneticField/(fabs(omtrack)*1000.));
    double recoThetaRad =  M_PI/2 - atan(tLtrack);
	  recoTheta.push_back(recoThetaRad*180./M_PI);
	  recoPhi.push_back(phtrack*180./M_PI);
	  recoD0.push_back(d0track);
	  recoZ0.push_back(z0track);
	  recoP.push_back(recoPt.back()/sin(recoThetaRad)); 
	  double tanrec = (double)tan((recoThetaRad*180./M_PI)/2);
	  double _recoEta = std::log(tanrec); 
	  recoEta.push_back(_recoEta);
	    double pttrackError = omtrackError/omtrack ;
	  _pttrackError.push_back(pttrackError*recoPt.back());
	  
	   resPt.push_back(recoPt.back()-truePt.back());
	   pullPt.push_back((recoPt.back()-truePt.back())/_pttrackError.back());//resPt_err
	   //pullPt.push_back(resPt.back()/pow(truePt.back(),2));
	   
	   
 recoNhits.push_back(track->getTrackerHits().size());
    if (track->getNdf()>0) recoChi2OverNDF.push_back(track->getChi2()/track->getNdf());
    else recoChi2OverNDF.push_back(-1.);
    
   
    /* else if(particlenum!=0)
    {
      
        for(int part=0; part<=particlenum; part++ )
         {
   
     MCParticle* particle = dynamic_cast<MCParticle*>(mcparticleVector.at(part));
    
    std::cout<<"This MCParticle PDG  "<<particle->getPDG()<<std::endl;
    
         
	 }
    
    } */
    
    
    }
     }
    perftree->Fill();
	// Increment the event number
	m_eventNumber++ ;
	delete relation;


}




void Global_residuals::check( LCEvent * evt ) {
  // nothing to check here - could be used to fill checkplots in reconstruction processor
}


void Global_residuals::end(){
  
  
	streamlog_out(MESSAGE) << " end()  " << name()
                         << " processed " << m_eventNumber << " events in " << m_runNumber << " runs "
                         << std::endl ;
  if (!m_useOnlyTree) {
   
    
    
    pulls = new TCanvas("pulls","Pulls of the track parameters",800,800);
    pulls->Divide(3,2);
    pulls->cd(1);
  }
  perftree->Write();
   //_tree->Write();
  _out->Write();
  _out->Close();

  
}


void Global_residuals::getCollection(LCCollection* &collection, std::string collectionName, LCEvent* evt){
  try{
    collection = evt->getCollection( collectionName ) ;
    cout<<"This collection names "<<evt->getDetectorName()<<endl;
    
  }
  catch(DataNotAvailableException &e){
    std::cout<<"- cannot get collections !!"<<std::endl;
    streamlog_out(DEBUG4) << "Collection " << collectionName.c_str() << " is unavailable" << std::endl;
    return;
  }
  return;
}
  

void Global_residuals::angleInFixedRange(double& angle){
  
  while (angle <= -M_PI ) angle = angle + 2*M_PI;
  while (angle >   M_PI ) angle = angle - 2*M_PI;
  
  return;
}

void Global_residuals::clearEventVariables(){
  
    truePt.clear();
  trueTheta.clear();
  truePhi.clear();
  trueD0.clear();
  trueZ0.clear();
  trueP.clear();
trueEta.clear();

  recoPt.clear();
  recoTheta.clear();
  recoPhi.clear();
  recoD0.clear();
  recoZ0.clear();
  recoP.clear();
recoEta.clear();

  recoNhits.clear();
  recoChi2OverNDF.clear();
  recoMinDist.clear();

  pullOm.clear();
  pullPh.clear();
  pullTL.clear();
  pullD0.clear();
  pullZ0.clear();
pullPt.clear();
  
  omegaResidual.clear();
  phiResidual.clear();
  resTanLambda.clear();
  resD0.clear();
  resZ0.clear();
resPt.clear(); 
//resPt_err.clear();

_d0trackError.clear();
  _phtrackError.clear();
  _omtrackError.clear();
_z0trackError.clear();
 _tLtrackError.clear();
_pttrackError.clear();

return;
  
}
