// Residuals code taken from Bogdan  
// Cleaned up - Sep 2016

#include "ResidualsCalculator.h" 
#include "TLatex.h"
#include "CLHEP/Vector/TwoVector.h"
#include "LinkDef.h"
#include "gear/CalorimeterParameters.h"
#include <cmath>
using namespace std ;
using namespace DD4hep ;
using namespace MarlinTrk ;
using namespace std;


ResidualsCalculator a_processor;

ResidualsCalculator::ResidualsCalculator() : Processor("ResidualsCalculator") {

	// Processor description
	_description = "ResidualsCalculator quantifies the tracking efficiency and produces residuals plots" ;

	// Output file
       /* registerProcessorParameter( "outFileNameR",
                                    "Name of the output root file",
				    _outFileName,
                                    std::string());
*/
	// Input collections
	registerInputCollection( LCIO::TRACK,
				 "TrackCollectionName",
				 "Track collection name",
				 m_inputTrackCollection,
				 std::string("SiTracks"));

        registerInputCollection( LCIO::MCPARTICLE,
                                 "MCParticleCollectionName",
                                 "Name of the MCParticle input collection",
                                 m_inputParticleCollection,
                                 std::string("MCParticle"));

				 		 registerInputCollection( LCIO::LCRELATION,
					"TrackRelationCollectionName",
                           "Track relation collection name",
                           m_inputTrackRelationCollection,	  
			 std::string("SiTrackRelations"));
	  
				 
	// Tracker hit collections
	std::vector<std::string> inputTrackerHitCollections;
	inputTrackerHitCollections.push_back(std::string("SiTrackerBarrelHitsPlanes"));
	inputTrackerHitCollections.push_back(std::string("SiTrackerEndcapHitsPlanes"));
	inputTrackerHitCollections.push_back(std::string("SiVertexBarrelHitsPlanes"));
	inputTrackerHitCollections.push_back(std::string("SiVertexEndcapHitsPlanes"));
	registerInputCollections( LCIO::TRACKERHITPLANE,
				  "TrackerHitCollectionNames",
				  "Name of the TrackerHit input collections",
				  m_inputTrackerHitCollections,
				  inputTrackerHitCollections);

	// Tracker hit relations collections
	std::vector<std::string> inputTrackerHitRelationCollections;
	inputTrackerHitRelationCollections.push_back(std::string("SiTrackerBarrelHitsRelations"));
	inputTrackerHitRelationCollections.push_back(std::string("SiTrackerEndcapHitsRelations"));
	inputTrackerHitRelationCollections.push_back(std::string("SiVertexBarrelHitsRelations"));
	inputTrackerHitRelationCollections.push_back(std::string("SiVertexEndcapHitsRelations"));
	registerInputCollections( LCIO::LCRELATION,
				  "TrackerHitRelCollectionNames",
				  "Name of TrackerHit relation collections",
				  m_inputTrackerHitRelationCollections,
				  inputTrackerHitRelationCollections);
	
	
				  
				  
	// Output collection
	registerOutputCollection( LCIO::MCPARTICLE,
                                  "MCParticleNotReco" ,
				  "Name of the output collection of the _not_reconstructed_ charged MCParticle",
				  m_notRecoMCColName,
                                  std::string("MCParticleNotReco"));

	// Process parameters
	registerProcessorParameter( "reconstructableDefinition",
				    "Set of cuts to define 'reconstractable' particles for efficiency computation. The options are: NHits, NHitsVXD, ILDLike",
				    m_cuts,
				    std::string("NHits"));
	
	registerProcessorParameter( "vertexBarrelID",
				    "Detector element ID for the vertex Barrel",
				    m_vertexBarrelID,
				    int(1));
	
}

void ResidualsCalculator::init() {
  // Print the initial parameters
  printParameters() ;
  
  // Reset counters
  m_runNumber = 0 ;
  m_eventNumber = 0 ;
	
  // Get the magnetic field
  DD4hep::Geometry::LCDD& lcdd = DD4hep::Geometry::LCDD::getInstance();
  const double position[3]={0,0,0}; // position to calculate magnetic field at (the origin in this case)
  double magneticFieldVector[3]={0,0,0}; // initialise object to hold magnetic field
  lcdd.field().magneticField(position,magneticFieldVector); // get the magnetic field vector from DD4hep
  m_magneticField = magneticFieldVector[2]/dd4hep::tesla; // z component at (0,0,0)
  
  // Define the purity cut - should be made configurable
  m_purity = 0.5;
  
  // Register this process
  Global::EVENTSEEDER->registerProcessor(this);
  
}
	
void ResidualsCalculator::processRunHeader( LCRunHeader* run) {
  ++m_runNumber ;
}


void ResidualsCalculator::processEvent( LCEvent* evt ) {

  if( isFirstEvent() ) {
    
    // Define plots
    eff_vs_theta = new TCanvas("eff_vs_theta","Tracks reconstruction Efficiency vs #Theta for SiD_o2_v01",800,800);
    g_eff_vs_theta = new TGraphAsymmErrors();
    
    // const int nbins_theta = 13;
    // double theta_edges[nbins_theta+1] = { 0.0, 10., 20., 30., 45., 60., 70, 87., 93., 110., 145., 160., 170., 180. };
    const int nbins_theta = 18;
    double theta_edges[nbins_theta+1] = { 0.0, 7., 15., 23., 30., 45., 60, 75, 88., 90., 92., 105., 120., 135., 150., 157., 165., 173., 180.};
    
    h_theta_reconstructed  = new TH1F( "h_theta_reconstructed", "Theta distributions of reconstructed tracks passing purity criteria", nbins_theta , theta_edges ) ;
    h_theta_reconstructable  = new TH1F( "h_theta_reconstructable", "Theta distribution of reconstructable tracks", nbins_theta , theta_edges );   
 
    eff_vs_pt = new TCanvas("eff_vs_pt","Trk Eff vs Pt",800,800);
    g_eff_vs_pt = new TGraphAsymmErrors() ;

    // const int nbins_pt = 12;
    // double pt_edges[nbins_pt+1] = { 0.01, 0.1, 0.5, 1., 2., 5., 10., 20., 50., 100., 200., 500., 1000. } ;
    const int nbins_pt = 17;//16  nbins 15 20 25 30 50 60 70 90 120 140 250 300 
    double pt_edges[nbins_pt+1] = { 0.01, 0.1, 0.2, 0.5, 1., 2., 4., 8., 10.,15., 40., 80.,100., 150., 200.,300., 500., 1000. } ;
    
    h_pt_reconstructed  = new TH1F( "h_pt_reconstructed", "Pt distributions of reconstructed tracks passing purity criteria", nbins_pt , pt_edges ) ;
    h_pt_reconstructable  = new TH1F( "h_pt_reconstructable", "Pt distribution of reconstructable tracks", nbins_pt , pt_edges ) ;
   
    h_pt_truth = new TH1F( "h_pt_truth", "Pt distributions of truth pions", 100 ,0.,50. ) ;
   
     h_pt_truth->GetXaxis()->SetTitle("Pt, GeV/c");
    
    h_pt_reco = new TH1F( "h_pt_rec", "Pt distributions of reconstructed pions ", 100 , 0.,50. ) ;
    
    h_pt_reco->GetXaxis()->SetTitle("Pt, GeV/c");
    
    
    h_energy = new TH1F( "h_energy", "Energy of pions ",100, 0. ,70.  ) ;
    
    h_energy->GetXaxis()->SetTitle("Energy, GeV");
    
    
    h_pt_res = new TH1F( "h_pt_residual", "Pt distributions of residuals ",100, -10. ,10.  ) ;
    
    
     h_particles = new TH1I( "h_particles", "Particles distributions by their PDG's ",100, -1000 ,1000  ) ;
     
     h_particles->GetXaxis()->SetTitle("PDG number");
  }
  
  std::cout << "Processing event " << m_eventNumber << std::endl;
  clearTreeVar();

  // First pick up all of the collections that will be used - tracks, MCparticles, hits from relevent subdetectors - and their relations
	
  //LCCollectionVec* skimVec = new LCCollectionVec( LCIO::MCPARTICLE )  ;
  //skimVec->setSubset(true) ; 

  // Initialise CELLID encoder to get the subdetector ID from a hit
  //UTIL::BitField64 m_encoder( lcio::ILDCellID0::encoder_string ) ;
  const string SiDecoderString="subdet:6,side:3,layer:4,module:12,sensor:1";
  UTIL::BitField64 m_encoder( SiDecoderString ) ;
  std::vector<int> m_collections;
  
  // Get the collection of tracks
  LCCollection* trackCollection = 0 ;
  getCollection(trackCollection, m_inputTrackCollection, evt); if(trackCollection == 0) return;
  
  // Get the collection of MC particles
  LCCollection* particleCollection = 0 ;
  getCollection(particleCollection, m_inputParticleCollection, evt); if(particleCollection == 0) return;
  
  // Get the collection of track relations to MC particles
  	LCCollection* trackRelationCollection = 0 ;
  	getCollection(trackRelationCollection, m_inputTrackRelationCollection, evt); if(trackRelationCollection == 0) return;
  
  // Create the relations navigator
 	LCRelationNavigator* trackRelation = new LCRelationNavigator( trackRelationCollection );
  
  // Make objects to hold all of the tracker hit and relation collections
  std::map<int,LCCollection*> trackerHitCollections;
  std::map<int,LCCollection*> trackerHitRelationCollections;
  std::map<int,LCRelationNavigator*> relations;
  std::map<int,bool> activeDetectors;
  
  // Loop over m_inputTrackerHitCollections and extract the data (only one now)
  for(unsigned int j=0; j < m_inputTrackerHitCollections.size(); j++){ 

    // Get the tracker hits
    LCCollection* trackerHitCollection = 0 ;
    getCollection(trackerHitCollection, m_inputTrackerHitCollections[j], evt); 

    cout << endl << "Collection " << j << ", "<< m_inputTrackerHitCollections[j] << ", has " << trackerHitCollection->getNumberOfElements() << " elements of type " << trackerHitCollection->getTypeName() << endl;
    
    if(trackerHitCollection == 0) continue;
    if(trackerHitCollection->getNumberOfElements() == 0) continue;

    // Get the tracker hit relations
    LCCollection* trackerHitRelationCollection = 0 ;
    getCollection(trackerHitRelationCollection, m_inputTrackerHitRelationCollections[j], evt);
    
    // Create the relations navigator
    LCRelationNavigator* relation = new LCRelationNavigator(trackerHitRelationCollection);
      
    // Get the subdetector for this set of collections
    int subdetector = getSubdetector(trackerHitCollection, m_encoder);
    cout << "This is THC subdetector array index" << subdetector << endl;

    // Fill arrays
    activeDetectors[subdetector] = true;
    trackerHitCollections[subdetector] = trackerHitCollection;
    relations[subdetector] = relation;
    m_collections.push_back(subdetector);
  }
  
  /*  LCCollection* trackRelationCollection = 0 ;
  getCollection(trackRelationCollection, m_inputTrackRelationCollection, evt); if(trackRelationCollection == 0) return;
  
	 LCRelationNavigator* relation = new LCRelationNavigator(trackerHitRelationCollection);
*/
  std::map<MCParticle*,int> particleTracks;
 int nTracks = trackCollection->getNumberOfElements();
	for(int itTrack=0;itTrack<nTracks;itTrack++){
	  // Get the track
    Track* track = dynamic_cast<Track*>( trackCollection->getElementAt(itTrack) ) ;
   std::cout << "This is D0 parameter of the track  " <<track->getD0() << std::endl;
	   const LCObjectVec& mcparticleVector = trackRelation->getRelatedToObjects( track );
	   MCParticle* particle = dynamic_cast<MCParticle*>(mcparticleVector.at(0));
	 //  particleTracks.insert(particle, 1);
	 particleTracks[particle] ++;
	 std::cout << " particleTracks have been written  " << std::endl;
	
	
	
	
	
	
	
	}
/*
  Now for each MC particle we want the list of hits belonging to it. The most
  efficient way is to loop over all hits once, and store the pointers in a
  map, with the key a pointer to the MC particle. We can then loop over each
  MC particle at the end and get all of the hits, before making a track.
*/
	
// Make the container
//std::map<MCParticle*, std::vector<TrackerHit*> > particleHits;
                            
 particleHits.clear();
 cout << endl << "Starting loop over subdetectors" << endl;
 for(unsigned int k=0; k < m_collections.size(); k++){
	
   // Get the subdetector number
   int subdetector = m_collections[k];
   cout << " subdetector "<< subdetector <<endl;
   
   // Find the hits
   int nHits = trackerHitCollections[subdetector]->getNumberOfElements();		
   cout << "Subdetector " << k << " has " << nHits << " hits" << endl;

   // Loop over tracker hits
   for(int itHit=0; itHit<nHits; itHit++){
     
     // Get the hit
     TrackerHitPlane* hit = dynamic_cast<TrackerHitPlane*>(trackerHitCollections[subdetector]->getElementAt(itHit)); //TrackerHitPlane
     cout<< "hit number "<< itHit <<endl;
     
     // Get the related simulated hit(s)
     const LCObjectVec& simHitVector = relations[subdetector]->getRelatedToObjects(hit);
     
     // Take the first hit only (this should be changed? Yes - loop over all related simHits and add an entry for each mcparticle so that this hit is in each fit)
     cout << "simHitVector size "<< simHitVector.size() <<endl;

     if(simHitVector.size()>0){
       SimTrackerHit* simHit = dynamic_cast<SimTrackerHit*>(simHitVector.at(0));	
       
       // Get the particle belonging to that hit
       MCParticle* particle = simHit->getMCParticle();
       
       // Push back the element into the container
       particleHits[particle].push_back(hit);
     }
			
   }
 }
 
 cout << "check1" << endl;

/*
  Now there is a map containing the number of tracks associated to each MC particle, and a 
  map of all hits on each particle. Loop over all MC particles, decide if they are 
  reconstructable or not, and check if they were reconstructed more than once (clones).
*/
 
 int nReconstructed=0;
 int nReconstructable=0;
 int nChargePart=0;
 int nCloseTrk=0;
 int nParticles = particleCollection->getNumberOfElements();
  
 
 LCRelationNavigator* relation = new LCRelationNavigator( trackRelationCollection );
 // Loop over MC particles
 for(int itParticle=0;itParticle<nParticles;itParticle++){
   nCloseTrk=0;
   // Get the particle
   MCParticle* particle = dynamic_cast<MCParticle*>( particleCollection->getElementAt(itParticle) ) ;
   // get theta angle for plot: eff vs theta
   
   
   
   std::cout<<"These are particles PDG's  in events  "<<particle->getPDG()<<std::endl;
   
    particles_distribution=particle->getPDG();
    
   if( particle->getPDG()==211 || particle->getPDG()==111  || particle->getPDG()==-211)//|| particle->getPDG()==-13 || particle->getPDG()==211 || particle->getPDG()==111  || particle->getPDG()==-211  particle->getPDG()==13  || particle->getPDG()==-13 
   {
     int nTracks = trackCollection->getNumberOfElements();
for(int itTrack=0;itTrack<nTracks;itTrack++){
  
  // Get the track
    Track* track = dynamic_cast<Track*>( trackCollection->getElementAt(itTrack) ) ;
    
    // Get the related list of MC particles
    const LCObjectVec& mcparticleVector = relation->getRelatedToObjects( track );
  
 TLorentzVector tv_mcp;
   tv_mcp.SetPxPyPzE(particle->getMomentum()[0],particle->getMomentum()[1],particle->getMomentum()[2],particle->getEnergy());
   //std::cout<<"This is distributed components of MCParticle  "<<std::setprecision(10)<< "[px]  "<< particle->getMomentum()[0]<<std::setprecision(10)<<"  [py]  "<<particle->getMomentum()[1]<<std::setprecision(10)<<"  [pz]  "<<particle->getMomentum()[2]<<" This is energy component  "<<std::setprecision(10)<<particle->getEnergy()<<std::endl ;
   double theta_mcp=tv_mcp.Theta()*180./M_PI;
   double pt_comp=sqrt(pow(particle->getMomentum()[0],2)+ pow(particle->getMomentum()[1],2)+pow(particle->getMomentum()[2],2) );
  // std::cout<<"This is the value of Truth P_t composite  "<<"This is just p  "<<pt_comp<<" pt "<<pt_comp*theta_mcp<<std::endl;
   //std::cout<<"This is distributed components of theta  "<<theta_mcp<<std::endl;
   double pt_mcp=tv_mcp.Pt();
    //  std::cout<<"This is the value of Truth P_t  "<<pt_mcp<<std::endl;
        HelixClass helix ;
    float pos[3]={particle->getVertex()[0],particle->getVertex()[1],particle->getVertex()[2]}; // Vertex position
    float mom[3]={particle->getMomentum()[0],particle->getMomentum()[1],particle->getMomentum()[2]}; // Particle momentum
    float charge = particle->getCharge() ; // Particle charge
    helix.Initialize_VP(pos,mom,charge,m_magneticField) ; // Initialised helix track
    
       ommcp = helix.getOmega() ;
      
       omtrack = track->getOmega() ;
      
     energy = particle->getEnergy() ;
     
      truePt=0.3*m_magneticField/fabs(ommcp*1000.);
      
      recoPt=0.3*m_magneticField/(fabs(omtrack)*1000.);
      
      resPt=(recoPt-truePt)/pow(truePt,2.);
      
      
    }
   bool isReco=false;
   // Check if it was reconstructed
   if(particleTracks.count(particle)){
     //std::cout<< "If statement have been passed  "<<std::endl;
     // Assumption: if particle was reconstructed then it is reconstructable!
     m_particles["all"]++; m_particles["reconstructable"]++; nReconstructable++;
     m_reconstructedParticles["all"]++; nReconstructed++;
     isReco=true;
     // Check if clones were produced (1 particle, more than 1 track)
     if(particleTracks[particle] > 1) m_reconstructedParticles["clones"]+=(particleTracks[particle]-1);
     
     //fill theta distribution for reconstructable mc particle (denominator of the eff) and for the reco one (numerator of the eff)
     h_theta_reconstructable -> Fill(theta_mcp);
     h_theta_reconstructed -> Fill(theta_mcp);
     
     h_pt_reconstructable -> Fill(pt_mcp);
     h_pt_reconstructed -> Fill(pt_mcp);
     
     h_pt_truth->Fill(truePt);
     h_pt_reco->Fill(recoPt);
     h_pt_res->Fill(resPt);
     h_energy->Fill(energy);
     
     h_particles->Fill(particles_distribution);
     
     continue;
   }
   }
   // Now decide on criteria for different particle types/classifications
   m_particles["all"]++; // all particles
   
   // No hits in the input collections
   if(particleHits.count(particle) == 0) continue; // No hits in the input collections
   
   // Exclude particles which are not "real"
   if(particle->vertexIsNotEndpointOfParent()) continue;
   
   if (!isReconstructable(particle,m_cuts)) continue;
   
   m_particles["reconstructable"]++; // reconstructable particles
   nReconstructable++;
   
   //fill theta distribution for reconstructable mc particle (denominator of the eff)
   h_theta_reconstructable->Fill(theta_mcp);
   
   h_pt_reconstructable->Fill(pt_mcp);   
 }

 cout<<"check2"<<endl;
  
 //evt->addCollection(  skimVec , m_notRecoMCColName ) ;
 
 struct sysinfo memInfo;
 sysinfo (&memInfo);
 
 unsigned long physMemUsed = memInfo.totalram - memInfo.freeram;
 physMemUsed *= memInfo.mem_unit; 
 streamlog_out(MESSAGE) << " Processed event  "<<m_eventNumber
			<< " Physical memory in use: "<<physMemUsed
			<< std::endl ;

 // Increment event number
 m_eventNumber++ ;
 std::cout << "For this event reconstructed " <<
   100.*(double)nReconstructed/(double)nReconstructable << 
   " % (" <<nReconstructed << "/" << nReconstructable << ")" << std::endl;
}


void ResidualsCalculator::check( LCEvent * evt ){
  // nothing to check here - could be used to fill checkplots in reconstruction processor
}	
 
void ResidualsCalculator::end(){
 // streamlog_out(MESSAGE) 
 std::cout<< " end()  " << name()
			 << " processed " << m_eventNumber << " events in " << m_runNumber << " runs "
			 << std::endl ;
  
  // Calculate efficiency results
  //streamlog_out(MESSAGE) 
 std::cout << std::fixed<<std::setprecision(2)<<"Reconstructable particle efficiency: " 
			 << 100.*m_reconstructedParticles["all"]/m_particles["reconstructable"]
			 << " % ("<<std::setprecision(0) << m_reconstructedParticles["all"] << "/" 
			 << m_particles["reconstructable"] << ")" << std::endl;
  
  
  // Plots
  eff_vs_theta->cd();

  g_eff_vs_theta->Divide( h_theta_reconstructed , h_theta_reconstructable , "v" ) ;
  g_eff_vs_theta->SetTitle("Tracks reconstruction Efficiency vs Theta for SiD_o2_v01");
  g_eff_vs_theta->SetMarkerColor(kRed) ;
  g_eff_vs_theta->SetLineColor(kRed) ;
  g_eff_vs_theta->GetYaxis()->SetTitle( "#epsilon_{trk}" );
  g_eff_vs_theta->GetXaxis()->SetTitle( "#theta [rad]" );
  
    TLatex*   tex4 = new TLatex(0.73,0.70,"p [15GeV;100GeV]");
  tex4->SetNDC();
  tex4->SetTextColor(6);
  tex4->SetTextFont(42);
  tex4->SetTextSize(0.056);
  tex4->SetLineWidth(2);
  tex4->Draw();
  
  
  
  g_eff_vs_theta->Draw("AP");
  g_eff_vs_theta->Write();
  eff_vs_theta->Write();

  h_theta_reconstructed->Write();
  h_theta_reconstructable->Write();

  eff_vs_pt->cd();
  gPad->SetLogx();

  g_eff_vs_pt->Divide( h_pt_reconstructed , h_pt_reconstructable , "v" ) ;
  g_eff_vs_pt->SetMarkerColor(kRed) ;
  g_eff_vs_pt->SetTitle("Tracks reconstruction Efficiency vs Pt for SiD_o1_v03");
  g_eff_vs_pt->SetLineColor(kRed) ;
  g_eff_vs_pt->GetYaxis()->SetTitle( "#epsilon_{trk}" );
  g_eff_vs_pt->GetXaxis()->SetTitle( "p_{T} [GeV]" );
  g_eff_vs_pt->Draw("AP");
  g_eff_vs_pt->Write();
  
        TLegend *leg = new TLegend(0.6,0.7,0.75,0.95);
  //leg->SetHeader("Polar Angles"); //name of the legend
  leg->AddEntry(g_eff_vs_theta,"p:[0,100]","p");
  //leg->AddEntry(Muon_plot20,"#theta = 20^{o}","p");
  //leg->AddEntry(g_eff_vs_pt,"#theta = 40^{o} single #mu","p");
  
  eff_vs_pt->Write();

  
  
  h_pt_reconstructed->Write();
  h_pt_reconstructable->Write();
  
  
  //_out->Write();
  //_out->Close();
  
  
}

void ResidualsCalculator::getCollection(LCCollection* &collection, std::string collectionName, LCEvent* evt){
  try{
    collection = evt->getCollection(collectionName) ;
  }
  catch(DataNotAvailableException &e){
    streamlog_out( DEBUG5 )<<"- cannot get collection. Collection " << collectionName.c_str() << " is unavailable" << std::endl;
    return;
  }
  return;
}


int ResidualsCalculator::getSubdetector(LCCollection* collection, UTIL::BitField64 &m_encoder){
  TrackerHitPlane* hit = dynamic_cast<TrackerHitPlane*>( collection->getElementAt(0)) ; 
  const int celId = hit->getCellID0();
 m_encoder.setValue(celId);
  
    const string SiDecoderString="subdet:6,side:3,layer:4,module:12,sensor:1";
  //UTIL::BitField64 m_encoder( SiDecoderString ) ;
  
  int subdet = m_encoder[lcio::ILDCellID0::layer];
  return subdet;
}

int ResidualsCalculator::getSubdetector(TrackerHit* hit, UTIL::BitField64 &m_encoder){
  
    const string SiDecoderString="subdet:6,side:3,layer:4,module:12,sensor:1";
  //UTIL::BitField64 m_encoder( SiDecoderString ) ;
  
  const int celId = hit->getCellID0() ;
  m_encoder.setValue(celId) ;
  int subdet = m_encoder[lcio::ILDCellID0::layer];//lcio::ILDCellID0::layer SiDecoderString SiDecoderString
  return subdet;
}

/*int ResidualsCalculator::getLayer(TrackerHit* hit, UTIL::Bitfield64 &m_encoder){

    const string SiDecoderString="subdet:6,side:3,layer:4,module:12,sensor:1";
 // UTIL::BitField64 m_encoder( SiDecoderString ) ;
  
  const int celId =hit->getCellID0();
encoder.setValue(celId);
int layer = m_encoder[ SiDecoderString]; 
cout<<"This layer value "<<layer<<endl;
return layer ;
  
}
*/
bool ResidualsCalculator::isReconstructable(MCParticle*& particle, std::string cut){
  
  if (cut=="NHits") {
    
    // Only make tracks with 6 or more hits
    std::vector<TrackerHit*> trackHits = particleHits[particle];
    if(trackHits.size() >= 6) return true;
    
  } else if (cut=="NHitsVXD") {
    
    // Only make tracks with 4 or more hits in the vertex detector
    std::vector<TrackerHit*> trackHits = particleHits[particle];
    int nVXDHits = 0;
    // new  modification 
    const string SiDecoderString="subdet:6,side:3,layer:4,module:12,sensor:1";
    UTIL::BitField64 m_encoder(SiDecoderString);
    for (size_t ihit=0; ihit<trackHits.size(); ihit++){
      int subdetector = getSubdetector(trackHits.at(ihit), m_encoder);
      if (subdetector == m_vertexBarrelID) nVXDHits++;
    }
    if (nVXDHits >= 4) return true;
    
  } else if (cut=="ILDLike") {
    
    // Only consider particles: charged, stable, pT>0.1GeV, cosTheta<0.89, nHits>=4, IP in 10 cm 
    //(a.t.m. cut in cosTheta is up to 0.89 instead of the usal 0.99 for cutting also particles in the vertex endcap region)
    //(a.t.m. no same cut on reco pT/theta but no bias because in that case also the mc particle is kept)
    
    //add condition for decay outside vertex/tracker? 
    bool isCharge = false;
    bool isStable = false;
    bool passPt = false;
    bool passTheta = false;
    bool passNHits = false;
    bool passIP = false;
    bool passEndPoint = false;

    double charge = fabs(particle->getCharge());
    if (charge>0.5) isCharge = true;
    //else std::cout<<"----- mc not charged"<<std::endl;
    int genStatus = particle->getGeneratorStatus();
    //int nDaughters = particle->getDaughters().size();
    if (genStatus == 1 ) isStable = true;
    //else std::cout<<"----- mc not stable in generator"<<std::endl;
    
    TLorentzVector p;
    p.SetPxPyPzE(particle->getMomentum()[0], particle->getMomentum()[1], particle->getMomentum()[2], particle->getEnergy());//in GeV
    if ( p.Pt()>=0.1 ) passPt = true;
    if ( fabs(cos(p.Theta()))<0.99 ) passTheta = true; 
    //if ( fabs(cos(p.Theta()))<0.8 ) passTheta = true; 
    //else std::cout<<"----- mc does not pass theta acceptance cut"<<std::endl;
    std::vector<TrackerHit*> trackHits = particleHits[particle];
    if(trackHits.size() >= 4) passNHits = true;
    else std::cout<<"----- mc has not 4 hits associated"<<std::endl;
    if(trackHits.size() >= 6) passNHits = true;
    
    int nVXDHits = 0;
    const string SiDecoderString="subdet:6,side:3,layer:4,module:12,sensor:1";
    UTIL::BitField64 m_encoder( SiDecoderString) ;
     for (size_t ihit=0; ihit<trackHits.size(); ihit++){
      int subdetector = getSubdetector(trackHits.at(ihit), m_encoder);
       if (subdetector == m_vertexBarrelID) nVXDHits++;
    }
     if (nVXDHits >= 4) return true;
    
    double dist = sqrt( pow(particle->getVertex()[0],2) + pow(particle->getVertex()[1],2) );
    if (dist<100.) passIP = true;
    //else std::cout<<"----- mc has dist from IP > 100mm"<<std::endl;
    //if (dist<30.) passIP = true;
    
    double e = sqrt( pow(particle->getEndpoint()[0],2) + pow(particle->getEndpoint()[1],2) );
    if (e==0. || e>40.) passEndPoint=true;
    //else std::cout<<"----- mc has endpoint < 40mm"<<std::endl;

    bool keepParticle = isCharge && isStable && passPt && passTheta && passNHits && passIP && passEndPoint;
    if (keepParticle) return true;
    
  } else {
    streamlog_out( ERROR )<<"Set of cuts " << cut.c_str() << " not defined" << std::endl;
    return false;
  }
  return false;
}

void ResidualsCalculator::clearTreeVar(){

  m_mcCat.clear();
  m_mcTheta.clear();
  m_mcPt.clear();
  m_mcTheta.clear();
  m_mcIsDecayedInTracker.clear();
  //m_mcNHits_helper.clear();
  //m_mcNHits.clear();
  m_mcNHitsTot.clear();
  m_mcNHitsVXD.clear();

  m_mcNTracks.clear();
  m_mcNTrkHits.clear();
  m_mcThetaTrk.clear();
  m_mcPtTrk.clear();
  m_mcPhiTrk.clear();
  m_mcNTracksCone.clear();
}  
